// Generated for Telelogic Tau 4.2.0.2.13102 at Thu Apr 16 19:56:53 MEST 2009
// Licensed Materials - Property of IBM 
//  Copyright IBM Corporation 2002,2008.  All Rights Reserved.
// 
// U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#if defined(TOR_USE_INSTRUMENTATION) && !defined(TOR_NO_EVENTMANAGER)

#include "torPlatform.h"

#include "tor.h"
#include "torTypes.h"
#include "torEventManager.h"
#include "torMetaEvents.h"
#include "torMetaStackContext.h"
#include "torEventManagerImpl.h" // Generated statemachine implementation
#include "torOs.h"
#include <assert.h>

namespace tor {

    namespace meta {

        // Representation of a pseudo-breakpoint, i.e. a breakpoint used internally by the eventManager in order
        // to implement debugger features such as "Break", "Step", "StepIn", "StepOut", "RunTo" etc.
        // N.B. Currently thread-focused pseudo-breakpoints are not supported.
        struct PseudoBreakpoint {
            // Returns true if the pseudo-breakpoint applies to the specified action when executed
            // by the specified thread. If it applies, execution should be halted at its location.
            // If the action already has been partially executed (in case the action contains operation calls)
            // the partiallyAlreadyExecuted flag should be set to true, otherwise false.
            virtual bool AppliesTo(const Charstring& actionGuid, 
                                   ThreadInfo& threadInfo,
                                   bool partiallyAlreadyExecuted) = 0;

            virtual ~PseudoBreakpoint() {}
        };

        // Break on the next action to execute for any thread
        struct BreakOnNextAction : PseudoBreakpoint {
            virtual bool AppliesTo(const Charstring& actionGuid, 
                                   ThreadInfo& threadInfo,
                                   bool partiallyAlreadyExecuted) {
                return true;
            }

            virtual ~BreakOnNextAction() {}
        };

        // Break on a specific action for a certain thread
        struct BreakOnSpecificAction : PseudoBreakpoint {
            BreakOnSpecificAction(const Charstring& actionGuid, int threadId) 
                : m_actionGuid(actionGuid), m_threadId(threadId) {}

                virtual bool AppliesTo(const Charstring& actionGuid, 
                                       ThreadInfo& threadInfo,
                                       bool partiallyAlreadyExecuted) {
                    if (threadInfo.m_id != m_threadId)
                        return false; // Not the correct thread

                    return (actionGuid == m_actionGuid);
                }

                virtual ~BreakOnSpecificAction() {}

                Charstring m_actionGuid;
                int m_threadId;
        };

        // Break on the next action to execute that fulfills a condition on the stack level
        // for a certain thread
        struct BreakOnNextActionWithStackCondition : PseudoBreakpoint {
            enum Condition {                
                SameOrOuter,    // Stack level must be less than or equal to the specified level (Step)
                Any,            // Any stack level allowed (StepIn)
                Outer           // Stack level must be less than the specified level (StepOut)
            };

            BreakOnNextActionWithStackCondition(unsigned int stackLevel, int threadId, Condition condition) 
                : m_stackLevel(stackLevel), m_threadId(threadId), m_condition(condition) {}

            virtual bool AppliesTo(const Charstring& actionGuid, 
                                   ThreadInfo& threadInfo,
                                   bool partiallyAlreadyExecuted) {
                if (threadInfo.m_id != m_threadId)
                    return false; // Not the correct thread

                if (m_condition == Any)
                    return true;

                unsigned int stackLevel = static_cast<unsigned int>(threadInfo.m_callstack.size());

                if (m_condition == SameOrOuter)
                    return (stackLevel <= m_stackLevel) && !partiallyAlreadyExecuted;

                assert(m_condition == Outer);
                return (stackLevel < m_stackLevel);
            }

            virtual ~BreakOnNextActionWithStackCondition() {}

            unsigned int m_stackLevel;
            int m_threadId;
            Condition m_condition;
        };    

    } // meta

} // tor

bool tor::meta::EventManager::receive(tor::Event* e) {
    // Just pass the event on to the event manager implementation.
    return emImpl->receive(e);
}

void tor::meta::EventManager::registerMedia(EventMedia* m) {
    os::auto_lock sync(m_data->m_mutex);

    emImpl->media.append(m);
}

void tor::meta::EventManager::unregisterMedia(EventMedia* m) {
    os::auto_lock lock(m_data->m_mutex);

    for (tor::Natural i = 1; i <= emImpl->media.length(); i = i + 1) {
        if (emImpl->media[i] == m)
        {
            emImpl->media.remove(i);
            break;
        }
    }
}

void tor::meta::EventManager::deleteAllLogFiles() {
    os::auto_lock lock(m_data->m_mutex);

    for (tor::Natural i = 1; i <= emImpl->media.length();) {
        EventMedia* media = emImpl->media[i];

        if (dynamic_cast<LogFile*>(media)) {
            emImpl->media.remove(i);
            delete media;
        }
        else
            i = i + 1;
    }
}

void tor::meta::EventManager::deleteAllHostTracers() {
    os::auto_lock lock(m_data->m_mutex);

    for (tor::Natural i = 1; i <= emImpl->media.length();) {
        EventMedia* media = emImpl->media[i];

        if (dynamic_cast<TauHostTracer*>(media)) {
            emImpl->media.remove(i);
            delete media;
        }
        else
            i = i + 1;
    }
}

void tor::meta::EventManager::createNewLogFile(const char* path) {
    tor::Charstring strPath( path );
    new LogFile(strPath);
}

void tor::meta::EventManager::createNewHostTracer() {
#if _MSC_VER
    new TauHostTracer();
#endif
}

void tor::meta::EventManager::setCoder(Coder* c) {
    os::auto_lock lock(m_data->m_mutex);

    coder = c;
}

tor::meta::Coder& tor::meta::EventManager::getDefaultCoder() {
    return emImpl->defaultCoder;
}

void tor::meta::EventManager::setActive(tor::Boolean a) {
    os::auto_lock lock(m_data->m_mutex);

    active = a;
}

bool tor::meta::EventManager::isActive() {
    if (isFinalized())
        return false; // When finalized the event manager cannot be active

    os::auto_lock lock(m_data->m_mutex);

    return active;
}

void tor::meta::EventManager::emitToMedia(tor::meta::Event* e) {
    for (tor::Natural i = 1; i <= emImpl->media.length(); i = i + 1) {
        emImpl->media[i]->emit(e->encodedInfo);
    }
}

bool tor::meta::EventManager::shouldWaitForAttach(tor::Charstring& host, tor::Natural& port) {
    host = _T("localhost");
    port = 56500;
    bool bShouldWaitForAttach;
    tor::Charstring hostDbgInfo;

#if defined(WIN32) && !defined(MAINWIN)
    size_t sz=0;
    getenv_s(&sz,NULL,0,"TTD_TOR_WAIT_FOR_ATTACH");

    bShouldWaitForAttach = (sz != 0);

    getenv_s(&sz,NULL,0,"TTD_TOR_HOST_DBG");
    char *t=NULL;
    if(sz>0)
    {
        t=new char[sz];
        getenv_s(&sz,t,sz,"TTD_TOR_HOST_DBG");
        hostDbgInfo=tor::Charstring(t);
        delete []t;
    }
#else
    bShouldWaitForAttach = (getenv("TTD_TOR_WAIT_FOR_ATTACH") != 0);

    char *host_dbg=getenv("TTD_TOR_HOST_DBG");
    hostDbgInfo=tor::Charstring(host_dbg?host_dbg:"");
#endif

    if (!hostDbgInfo.stringRep.empty()) {
        Charstring::StringRepType::size_type s = hostDbgInfo.stringRep.find_last_of(_T("#"));
        if ((s != Charstring::StringRepType::npos) && (s != 0) && (s != hostDbgInfo.stringRep.size() - 1)) {
            host = hostDbgInfo.stringRep.substr(0, s);
            Charstring strPort(hostDbgInfo.stringRep.substr(s+1));
            port = atoi(strPort.getNarrowString().c_str());
        }
    }

    return bShouldWaitForAttach;
}

void tor::meta::EventManager::removeAllPseudoBreakpoints() {
    os::auto_lock lock(m_data->m_mutex);

    if (m_data->m_pseudoBreakpoint) {
        delete m_data->m_pseudoBreakpoint;
        m_data->m_pseudoBreakpoint = 0;
    }
}

void tor::meta::EventManager::continueApplication(tor::Event* event) {
    os::auto_lock lock(m_data->m_mutex);

    tor::meta::Event* e = cast<tor::meta::Event*>(event);
    if (!e) {
        assert(0); // Should only process meta events here...
        return;
    }

    if (!e->m_bSynchronouslySent)
        return; // Nothing to do. The application is not waiting for an ack.

    for (ThreadInfos::iterator it = emImpl->threadInfos.begin();
        it != emImpl->threadInfos.end(); it++) 
    {
        ThreadInfo* ti = (*it).second;
        ti->m_gate.unlock();
    }
}

tor::Charstring& tor::meta::EventManager::getCurrentExecutionPoint() {
    os::auto_lock lock(m_data->m_mutex);

    ThreadInfo* threadInfo = _getActiveThread();
    assert(threadInfo);
    if (!threadInfo) {
        static Charstring error(_T("Interal Error! No active threads!")); // HCS
        return error;
    }

    StackFrame& stackFrame = threadInfo->m_callstack.top();
    return stackFrame.m_actionGuid;
}

tor::meta::ThreadInfo* tor::meta::EventManager::_getActiveThread() {
    if (emImpl->threadInfos.size() == 0)
        return 0; // Strange. There should at least be one user thread.

    ThreadInfo* threadInfo = 0;
    for (ThreadInfos::iterator it = emImpl->threadInfos.begin();
        it != emImpl->threadInfos.end(); it++) 
    {
        ThreadInfo* ti = (*it).second;
        if (ti->m_active) {
            threadInfo = ti;
            break;
        }
    }

    if (!threadInfo) {
        // No thread is in the active debug context. Make the first one active.
        ThreadInfos::iterator it = emImpl->threadInfos.begin();
        ThreadInfo* ti = (*it).second;
        ti->m_active = true;
        threadInfo = ti;            
    }

    return threadInfo;
}

void tor::meta::EventManager::onStep() {
    os::auto_lock lock(m_data->m_mutex);

    if (!m_data->m_pseudoBreakpoint) {
        // N.B. This function is always called from the context of the eventManager thread.
        // Locate one of the user threads that is in the active debug context and set a pseudobreakpoint on it...        

        ThreadInfo* threadInfo = _getActiveThread();
        if (!threadInfo)
            return;

        unsigned int stackLevel = static_cast<unsigned int>(threadInfo->m_callstack.size());
        m_data->m_pseudoBreakpoint = new BreakOnNextActionWithStackCondition(stackLevel, 
            threadInfo->m_id,
            BreakOnNextActionWithStackCondition::SameOrOuter);
    }
}

void tor::meta::EventManager::onStepIn() {
    os::auto_lock lock(m_data->m_mutex);

    if (!m_data->m_pseudoBreakpoint) {
        // N.B. This function is always called from the context of the eventManager thread.
        // Locate one of the user threads that is in the active debug context and set a pseudobreakpoint on it...        

        ThreadInfo* threadInfo = _getActiveThread();
        if (!threadInfo)
            return;

        unsigned int stackLevel = static_cast<unsigned int>(threadInfo->m_callstack.size());
        m_data->m_pseudoBreakpoint = new BreakOnNextActionWithStackCondition(stackLevel, 
            threadInfo->m_id,
            BreakOnNextActionWithStackCondition::Any);
    }
}

void tor::meta::EventManager::onStepOut() {
    os::auto_lock lock(m_data->m_mutex);

    if (!m_data->m_pseudoBreakpoint) {
        // N.B. This function is always called from the context of the eventManager thread.
        // Locate one of the user threads that is in the active debug context and set a pseudobreakpoint on it...        

        ThreadInfo* threadInfo = _getActiveThread();
        if (!threadInfo)
            return;

        unsigned int stackLevel = static_cast<unsigned int>(threadInfo->m_callstack.size());
        m_data->m_pseudoBreakpoint = new BreakOnNextActionWithStackCondition(stackLevel, 
            threadInfo->m_id,
            BreakOnNextActionWithStackCondition::Outer);
    }
}

void tor::meta::EventManager::onRunTo(const tor::Charstring& actionGuid) {
    os::auto_lock lock(m_data->m_mutex);

    if (!m_data->m_pseudoBreakpoint) {
        // N.B. This function is always called from the context of the eventManager thread.
        // Locate one of the user threads that is in the active debug context and set a pseudobreakpoint on it...        

        ThreadInfo* threadInfo = _getActiveThread();
        if (!threadInfo)
            return;

        m_data->m_pseudoBreakpoint = new BreakOnSpecificAction(actionGuid, threadInfo->m_id);
    }
}

void tor::meta::EventManager::onBreak() {
    os::auto_lock lock(m_data->m_mutex);

    if (m_data->m_pseudoBreakpoint)
        // There was already a pseudo breakpoint set. Delete it, since Break has higher priority
        // (It shall always be possible to Break the program while running.).
        delete m_data->m_pseudoBreakpoint;
        
    m_data->m_pseudoBreakpoint = new BreakOnNextAction;
}

void tor::meta::EventManager::setBreakpoints(String<BreakpointInfo>& breakpointsToSet) {
    os::auto_lock lock(m_data->m_mutex);

    for (String<BreakpointInfo>::iterator it = breakpointsToSet.begin(); 
        it != breakpointsToSet.end(); it++) {
            BreakpointInfo& bpInfo = *it;
            m_data->m_breakpointMap.insert(BreakpointMap::value_type(bpInfo.actionGuid, bpInfo));
        }
}

void tor::meta::EventManager::removeBreakpoints(String<BreakpointInfo>& breakpointsToRemove) {
    os::auto_lock lock(m_data->m_mutex);

    if (breakpointsToRemove.empty()) {
        // Remove all breakpoints
        m_data->m_breakpointMap.clear();
        return;
    }

    for (String<BreakpointInfo>::iterator it = breakpointsToRemove.begin(); 
        it != breakpointsToRemove.end(); it++) {
            BreakpointInfo& bpInfo = *it;
            m_data->m_breakpointMap.erase(bpInfo.actionGuid);
        }
}

tor::meta::AppInfo tor::meta::EventManager::getAppInfo() {
    // Currently the application information only includes the GUID of the build artifacts that was used when building 
    // the application.
    AppInfo appInfo(theApplication->guid);
    return appInfo;
}

void tor::meta::EventManager::processEvent(tor::meta::Event* event) {
    if (!isActive())
        return;

    int threadId = os::Thread::id();
    ThreadInfo* threadInfo = getThreadInfo(threadId);

    // Different events require different handling here...
    if (cast<BreakpointHit_Charstring*>(event)) {
        // The current thread has reached a breakpoint. Block all other threads until execution
        // shall be resumed.
        blockAllThreadsExceptOne(threadId);
    }

    event->encode(coder); // To get a string representation of the event (stored in the event itself)
    if (event->m_bSynchronouslySent) {
        threadInfo->m_gate.lock();
        tor::sendTo(event, emImpl);
        threadInfo->m_gate.enter(); // Wait until the event has been processed (then the gate is opened)
    }
    else 
        // Asynchronous sending of the event
        tor::sendTo(event, emImpl);
}

static int eventManagerInitCount = 0;

tor::meta::ThreadInfo* tor::meta::EventManager::getThreadInfo(int threadId) {
    os::auto_lock lock(m_data->m_mutex);
    return _getThreadInfo(threadId);
}

tor::meta::ThreadInfo* tor::meta::EventManager::_getThreadInfo(int threadId) {
    ThreadInfos::const_iterator it = emImpl->threadInfos.find(threadId);
    if (it != emImpl->threadInfos.end())
        return (*it).second;

    ThreadInfo* threadInfo = new ThreadInfo(threadId);
    emImpl->threadInfos.insert(ThreadInfos::value_type(threadId, threadInfo));

    return threadInfo;
}

void tor::meta::EventManager::blockAllThreadsExceptOne(int threadId) {
    os::auto_lock lock(m_data->m_mutex);

    for (ThreadInfos::iterator it = emImpl->threadInfos.begin();
        it != emImpl->threadInfos.end(); it++) 
    {
        ThreadInfo* ti = (*it).second;
        if (ti->m_id != threadId)
            ti->m_gate.lock(); // Block the thread by locking its gate
    }
}

bool tor::meta::EventManager::isFinalized() {
    return eventManagerInitCount == -1;
}

void tor::meta::EventManager::finalize() {
    // Explicit immediate finalization!
    eventManagerInitCount = 0;
    eventManager.EventManager::~EventManager();
    eventManagerInitCount = -1;
}

void tor::meta::EventManager::initialize() {
    // Explicit immediate initialization!
    eventManagerInitCount = 0;
    new (&eventManager) EventManager();   
    eventManagerInitCount = 1;
}

void tor::meta::EventManager::beforeAction(const Charstring& strGuid, bool partiallyAlreadyExecuted /* = false */) {
    if (isFinalized())
        // The eventManager is finalized (probably explicitly finalized). Do nothing.
        return;

    bool bBreakpointHit = false;

    { // Start synchronized region

        os::auto_lock lock(m_data->m_mutex);

        int threadId = os::Thread::id();
        ThreadInfo* threadInfo = _getThreadInfo(threadId);

        if (threadInfo->m_callstack.empty()) {
            // This should normally not happen! How could we execute an action without having a callstack!?!
            // However, one explanation could be that TOR was just initialized so that no stack frames
            // exist yet. 'beforeAction' will only have an effect after at least one stack frame exist.
            // Another explanation could be that the generated instrumentation was modified by the user.
            return; 
        }

        // Set up information about current execution point
        StackFrame& stackFrame = threadInfo->m_callstack.top();
        stackFrame.m_actionGuid = strGuid;

        // Check if we should break on this action.

        // First check for applicable pseudo-breakpoint
        if (m_data->m_pseudoBreakpoint) {            
            if (m_data->m_pseudoBreakpoint->AppliesTo(strGuid, *threadInfo, partiallyAlreadyExecuted)) 
                // Pseudo-breakpoint hit
                bBreakpointHit = true;
        }

        if (!bBreakpointHit) {
            // No pseudo-breakpoint! Next check for applicable ordinary breakpoint
            BreakpointMap::const_iterator it = m_data->m_breakpointMap.find(strGuid);
            if (it != m_data->m_breakpointMap.end()) {
                // Ordinary breakpoint hit.
                bBreakpointHit = true;

                // Make the current thread part of the active debug context (i.e. we now start to debug that thread if not already done)
                threadInfo->m_active = true;
            }
        }

    } // End synchronized region

    if (bBreakpointHit) {
        // N.B. No signal sending within synchronized region, or we will be dead-locked!
        processEvent(new BreakpointHit_Charstring(strGuid));
    }
    else {
        // Finally, check if the current thread shall be blocked. If so, block it now.
        int threadId = os::Thread::id();
        ThreadInfo* threadInfo = getThreadInfo(threadId);
        threadInfo->m_gate.enter();
    }
}

tor::meta::EventManager tor::meta::eventManager;

// N.B. This constructor is executed more than once! Once from the placement new in the EventManagerInit
// and once from the normal construction of the eventManager. Thus all initialization that must not happen
// twice must be within the constructor code block!
tor::meta::EventManager::EventManager() {
    if (eventManagerInitCount > 0)
        return; // Event manager already initialized!

    active = true;
#ifdef TOR_SYNC_META_EVENTS
    synchronous = true;
#else
    synchronous = false;
#endif
    m_data = new EventManagerData;
    emImpl = new EventManagerImpl;
    coder = &emImpl->defaultCoder;
}

tor::meta::EventManager::~EventManager() {
    if (eventManagerInitCount != 0) 
        return; // Destruction must not take place while the EventManager is still initialized,
    // or if has already been finalized.

    if (coder && (coder != &emImpl->defaultCoder))
        delete coder;

    if (emImpl)
        delete emImpl;   

    if (m_data)
        delete m_data;

    eventManagerInitCount = -1; // Don't finalize again!
}

tor::meta::EventManagerInit::EventManagerInit() { 
    if (eventManagerInitCount == 0) {
        // Initialize the event manager
        new (&eventManager) EventManager();
    }
    eventManagerInitCount++;
}

tor::meta::EventManagerInit::~EventManagerInit() { 
    eventManagerInitCount--;
    if (eventManagerInitCount == 0) {
        // Free event manager memory
        eventManager.EventManager::~EventManager();
    }
}

// Include generated implementation
#include "torEventManagerImpl.cpp"

//////////////////////////////////////////////////////////////////////////


void tor::trace(const tor::Charstring& message) {
    tor::meta::eventManager.processEvent(new tor::meta::Message(message));
}

void tor::assertFailure(const TCHAR* condition, const TCHAR* file, int line) {
    tor::Charstring str(_T("Assertion failed!\nCondition: "));
    str += condition;
    str += _T("\nSource file: ");
    str += file;
    str += _T("\nLine: ");
    TCHAR buf[16];
    stprintf_number(buf, line);    
    str += buf;
    str += _T("\n");

    trace(str);
}


#endif 

